import { Effect } from "effect";
import { instantiateRefEncoded, isRefEncoded, subscriptionsScopes, } from "../internal.js";
export class Ref {
    constructor(id, controlledAccount, schema) {
        this.id = id;
        this.controlledAccount = controlledAccount;
        this.schema = schema;
        if (!isRefEncoded(schema)) {
            throw new Error("Ref must be constructed with a ref schema");
        }
    }
    get value() {
        if (this.cachedValue)
            return this.cachedValue;
        // TODO: cache it for object identity!!!
        const raw = this.controlledAccount._raw.core.node.getLoaded(this.id);
        if (raw) {
            const value = instantiateRefEncoded(this.schema, raw);
            this.cachedValue = value;
            return value;
        }
        else {
            return null;
        }
    }
    loadEf() {
        return Effect.async((fulfill) => {
            this.loadHelper()
                .then((value) => {
                if (value === "unavailable") {
                    fulfill(Effect.fail("unavailable"));
                }
                else {
                    fulfill(Effect.succeed(value));
                }
            })
                .catch((e) => {
                fulfill(Effect.die(e));
            });
        });
    }
    async loadHelper(options) {
        const raw = await this.controlledAccount._raw.core.node.load(this.id, options?.onProgress);
        if (raw === "unavailable") {
            return "unavailable";
        }
        else {
            return new Ref(this.id, this.controlledAccount, this.schema).value;
        }
    }
    async load(options) {
        const result = await this.loadHelper(options);
        if (result === "unavailable") {
            return undefined;
        }
        else {
            return result;
        }
    }
    accessFrom(fromScopeValue) {
        const subScope = subscriptionsScopes.get(fromScopeValue);
        subScope?.onRefAccessedOrSet(this.id);
        if (this.value && subScope) {
            subscriptionsScopes.set(this.value, subScope);
        }
        return this.value;
    }
}
export function makeRefs(getIdForKey, getKeysWithIds, controlledAccount, refSchemaForKey) {
    const refs = {};
    return new Proxy(refs, {
        get(_target, key) {
            if (key === Symbol.iterator) {
                return function* () {
                    for (const key of getKeysWithIds()) {
                        yield new Ref(getIdForKey(key), controlledAccount, refSchemaForKey(key));
                    }
                };
            }
            if (typeof key === "symbol")
                return undefined;
            if (key === "length") {
                return getKeysWithIds().length;
            }
            const id = getIdForKey(key);
            if (!id)
                return undefined;
            return new Ref(id, controlledAccount, refSchemaForKey(key));
        },
        ownKeys() {
            return getKeysWithIds().map((key) => key.toString());
        },
        getOwnPropertyDescriptor(target, key) {
            const id = getIdForKey(key);
            if (id) {
                return {
                    enumerable: true,
                    configurable: true,
                    writable: true,
                };
            }
            else {
                return Reflect.getOwnPropertyDescriptor(target, key);
            }
        },
    });
}
//# sourceMappingURL=refs.js.map